# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import tempfile

from IPython.core.magic import Magics, line_cell_magic, line_magic, magics_class


class HistoryContext:
    def __init__(self):
        self.history = []
        self.history_start = 1
        self.history_end = 1

    def get_history(self):
        return self.history

    def get_history_without_ipython_calls(self):
        """
        Remove all lines in execution history added by iPython calls.
        """
        line_count = 0
        filtered_history = []
        for line in self.history:
            if "get_ipython()." not in line:
                filtered_history.append(line)
                line_count += 1
        return filtered_history, line_count

    def add_to_history(self, lines):
        self.history += lines

    def get_history_start(self):
        return self.history_start

    def get_history_end(self):
        return self.history_end

    def increment_end(self):
        self.history_end += 1

    def record_history_write(self):
        self.history_start = self.history_end

    def reset_history(self):
        self.history = []
        self.history_start = self.history_end


history_context = HistoryContext()


def process_input():
    history_context.increment_end()


def call_client(source_data, error_start):
    with tempfile.TemporaryFile("w+") as source_file:
        source_file.write("\n".join(source_data))
        source_file.seek(0)  # for debugging
        print(source_file.read())  # for debugging
    print("Start returning errors on line {}".format(error_start))  # for debugging


@magics_class
class ExecutePyre(Magics):
    """
    Usage example
    ---------
    In [1]: %load_ext pyre_ipython
    Pyre execution magic loaded.
    In [2]: %%check
       ...: def foo():
       ...:     return x
    In [3]: %check x = 1
    In [4]: check x = 1
    """

    @line_cell_magic
    def check(self, line, cell=None):
        """
        Typecheck input and add it to the environment.
        """
        code = cell or line
        ipython = get_ipython()  # noqa
        self.write_history()

        # create sourcefile and perform typecheck
        source_data, line_count = history_context.get_history_without_ipython_calls()
        code_lines = code.split("\n")
        source_data += code_lines
        call_client(source_data, line_count + 1)

        # add checked code to history
        history_context.add_to_history(code_lines)
        ipython.run_code(code)

    @line_cell_magic
    def check_no_exec(self, line, cell=None):
        """
        Typecheck input without executing it or adding it to the future
        environment.
        """
        code = cell or line
        ipython = get_ipython()  # noqa
        self.write_history()

        # create sourcefile and perform typecheck
        source_data, line_count = history_context.get_history_without_ipython_calls()
        source_data += code.split("\n")
        call_client(source_data, line_count + 1)

    @line_magic
    def env(self, line=None):
        """
        Display the current type environment.
        """
        self.write_history()
        print(history_context.get_history())  # logging (to be deleted)

    @line_magic
    def clear_env(self, line=None):
        """
        Clear the type environment and all previously executed history.
        """
        history_context.reset_history()
        print("Environment has been reset.")

    def write_history(self):
        temp_file = "tempfile.py"
        history_command = "history {}-{} -t -f {}".format(
            history_context.get_history_start(),
            history_context.get_history_end(),
            temp_file,
        )
        if os.path.exists(temp_file):
            os.remove(temp_file)
        ipython = get_ipython()  # noqa
        ipython.magic(history_command)

        lines = []
        with open(temp_file, "r") as temp_file_readable:
            lines = temp_file_readable.readlines()
        lines_list = [line.strip() for line in lines]
        history_context.add_to_history(lines_list)
        history_context.record_history_write()
        os.remove(temp_file)


def load_ipython_extension(ipython):
    ipython.register_magics(ExecutePyre)
    ipython.events.register("pre_execute", process_input)

    existing_history = "existing_history.py"
    history_command = "history -f {}".format(existing_history)
    ipython.magic(history_command)
    with open(existing_history, "r") as existing_history_readable:
        for _i, _l in enumerate(existing_history_readable):
            history_context.increment_end()
    history_context.reset_history()
    os.remove(existing_history)
    print("Pyre execution magic loaded.")


def unload_ipython_extension(ipython):
    history_context.reset_history()



from typing import Any, Dict, List, Optional, Tuple, Type

_defaultaction: str
_onceregistry: Dict[Any, Any]
filters: List[Tuple[Any, ...]]

def warn(message: Warning, category: Optional[Type[Warning]] = ..., stacklevel: int = ...) -> None: ...
def warn_explicit(message: Warning, category: Optional[Type[Warning]],
                  filename: str, lineno: int,
                  module: Any = ..., registry: Dict[Any, Any] = ...,
                  module_globals: Dict[Any, Any] = ...) -> None: ...



# Stubs for maxminddb (Python 3)

from typing import Any

from maxminddb import reader

def open_database(database: str, mode: int = ...) -> reader.Reader: ...

def Reader(database: str) -> reader.Reader: ...



import glob
import os
import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 5):
    sys.exit('Error: pyre-check only runs on Python 3.5 and above.')

def find_typeshed_files(base):
    if not os.path.isdir(base):
       return []
    typeshed_root = os.path.join(base, 'typeshed')
    if not os.path.isdir(typeshed_root):
       return []
    result = []
    for absolute_directory, _, _ in os.walk(typeshed_root):
        relative_directory = os.path.relpath(absolute_directory, base)
        files = glob.glob(os.path.join(relative_directory, '*.pyi'))
        if not files:
            continue
        target = os.path.join('lib', 'pyre_check', relative_directory)
        result.append((target, files))
    return result

with open('README.md') as f:
    long_description = f.read()


setup(
    name='pyre-check',
    version='0.0.36',
    description='A performant type checker for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://pyre-check.org/',
    download_url='https://github.com/facebook/pyre-check',
    author='Facebook',
    author_email='pyre@fb.com',
    maintainer='Facebook',
    maintainer_email='pyre@fb.com',
    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
    ],
    keywords='typechecker development',

    packages=find_packages(exclude=['tests', 'pyre-check']),
    data_files=[('bin', ['bin/pyre.bin'])] + find_typeshed_files("/tmp/tmp.L6ffURTABV/"),
    python_requires='>=3.5',
    install_requires=['pywatchman', 'psutil', 'libcst', 'pyre_extensions'],
    entry_points={
        'console_scripts': [
            'pyre = pyre_check.client.pyre:main',
            'pyre-upgrade = pyre_check.tools.upgrade.upgrade:main',
        ],
    }
)
